/*#define PROCDEBUG*/
/*
**  PostNews.c
**
**  Copyright 1988-1991 by Matthew Dillon, All Rights Reserved.
**  Extensive Changes Copyright 1993 by Michael B. Smith. All Rights Reserved.
**
**  POSTNEWS <newsfile [-f from-username] [-r from-realname]
**	     [-R references-file] [-x delfile] [newsfile]
**
**  PostNews takes "newsfile", which may either be the LAST argument,
**  or it may be stdin, and attempts to post it.
**
**  The "references-file" is typically generated by a newsreader, and
**  contains the references list that should be included in the headers
**  of the outgoing article. Regardless, when specified, the full
**  contents of "references-file" are placed into the headers of the
**  article. Any empty lines in "references-file" can cause a bad
**  news article to occur.
**
**  Files listed after "-x" (there may be upto 31) are deleted if the
**  article is posted successfully. These are typically temporary files
**  created by a newsreader, and it doesn't want to wait for PostNews
**  to complete to clean up.
**
**  The "from-username" and the "real-username" are used for generating
**  the article's "From:" header, if it doesn't already exist in the
**  input files ("newsfile" and "references-file" are both checked).
**
**  If not specified, obtains user/realname from USERNAME and REALNAME
**  environment variables.  If either does not exist then gets it from
**  the UserName and RealName config entries.
**
**  A username, a realname, and a nodename *must* exist to post news.
**
**  If a Newsgroups header cannot be found, news is not posted.
**
**  If any of "From", "Date", "Message-Id", "Organization" or "Subject"
**  headers are not supplied, PostNews will generate one.
*/

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "config.h"
#include "version.h"
#include "owndevunit.h"
#include "log.h"

#define PROG_NAME    "PostNews"

IDENT (".10");

/*
**  Globals
*/

int
	_bufsiz = 8192,     /* stdio buffer size */
	Di = 0;
char
	TmpBuf [2048],
	*DelFile [32];
struct Library
	*OwnDevUnitBase = NULL;

/*
**  Forwards
*/

int  PostNews (FILE *, FILE *, char *, char *);
void myexit (void);

/*
**  Debugging help
*/

#ifdef PROCDEBUG
#define PROC(x)     const char *_proc = x; printf ("%s: enter\n", _proc);
#define D(x)	    printf ("%s: ", _proc); printf x
#else
#define PROC(x)
#define D(x)
#endif

void
err (char *msg1, char *msg2)
{
	PROC ("err");

	D ((msg1, msg2));

	if (msg1 || msg2) {
		fprintf (stderr, msg1, msg2);
		fprintf (stderr, "\n");
	}

	ulog (-1, msg1, msg2);

	D (("exit\n"));
	return;
}

void fatal (char *msg1, char *msg2)
{
	PROC ("fatal");

	err (msg1, msg2);

	D (("exit 20\n"));
	exit (20);
}

int
main (int ac, char **av)
{
	int
		i;
	char
		*from = NULL,
		*real = NULL,
		*reffile = NULL;
	FILE
		*rfi = NULL;

	PROC ("main");

	LogProgram = PROG_NAME;

	atexit (myexit);

	if ((OwnDevUnitBase = OpenLibrary (ODU_NAME, 0)) == NULL) {
		fatal ("Unable to open '%s'", ODU_NAME);
	}

	for (i = 1; i < ac; ++i) {
		char
			*ptr = av [i];

		if (*ptr == '-') {
			ptr += 2;
			switch (ptr [-1]) {
				case 'f':
					if (*ptr == 0)
						ptr = av [++i];
					from = ptr;
					break;
				case 'r':
					if (*ptr == 0)
						ptr = av [++i];
					real = ptr;
					break;
				case 'R':
					if (*ptr == 0)
						ptr = av [++i];
					reffile = ptr;
					break;
				case 'x':
					if (*ptr == 0)
						ptr = av [++i];
					DelFile [Di++] = ptr;
					if (Di >= 32) {
						fatal ("Too many files to delete", NULL);
					}
					break;
				default:
					fatal ("Unknown option: '%s'", ptr - 2);
			}
		}
		else {
			if (freopen (ptr, "r", stdin) == NULL) {
				fatal ("cannot open '%s'", ptr);
			}
		}
	}

	if (i > ac) {
		err ("Invalid argument list\n", NULL);
	}

	if (from == NULL)
		from = GetUserName ();
	if (real == NULL)
		real = GetRealName ();
	if (from == NULL || real == NULL) {
		fatal ("Incomplete configuration, can't post news.\n", NULL);
	}

	if (reffile) {
		rfi = fopen (reffile, "r");
		D (("reffile '%s' (0x%lx)\n", reffile, rfi));
	}

	LockFile ("PostNews-UPDATE");
	i = PostNews (stdin, rfi, from, real);
	UnLockFile ("PostNews-UPDATE");

	D (("PostNews rslt %ld\n", i));

	if (rfi)
		fclose (rfi);
	fclose (stdin);

	if (i == 0) {
		/* don't remove the files if posting failed */
		while (--Di >= 0) {
			D (("removing '%s'\n", DelFile [Di]));
			remove (DelFile [Di]);
		}
	}
	else {
		D (("exit 20\n"));
		exit (20);
	}

	D (("exit 0\n"));
	exit (0);
}

void
myexit (void)
{
	PROC ("myexit");

	UnLockFiles ();

	if (OwnDevUnitBase) {
		CloseLibrary (OwnDevUnitBase);
		OwnDevUnitBase = NULL;
	}

	D (("exit\n"));
	return;
}

/* STRIEQN is an optimized strnicmp (a, b) == 0; assume non-NULL */
/*	   for both; also assumes that 'b' is a "const char []"  */
/*	   (idea stolen from C News. Thanks Henry and Geoff!)	 */
#define STRIEQN(a, b) (toupper ((a) [0]) == toupper ((b) [0]) && strnicmp (a, b, sizeof (b) - 1) == 0)

int
PostNews (FILE *fi, FILE *rfi, char *user, char *realname)
{
	char
		*domain       = FindConfig (DOMAINNAME),
		*nodename     = FindConfig (NODENAME),
		*timezone     = FindConfig (TIMEZONE),
		*autoBatch    = FindConfig (AUTOBATCH),
		*organization = FindConfig (ORGANIZATION),
		*rnewsnop     = FindConfig ("RNewsNoP"),
		*rnews	      = GetConfigProgram (RNEWS);
	int
		firsttime,
		rslt,
		seq = GetSequence (2);
	char
		tfname [35];
	FILE
		*fo;
	const char
		HFrom	    [] = "From: ",
		HDate	    [] = "Date: ",
		HMsgid	    [] = "Message-Id: ",
		HOrg	    [] = "Organization: ",
		HPath	    [] = "Path: ",
		HSubj	    [] = "Subject: ",
		HNewsg	    [] = "Newsgroups: ";
	int
		hFrom	    = 0,
		hDate	    = 0,
		hMsgid	    = 0,
		hOrg	    = 0,
		hPath	    = 0,
		hSubj	    = 0,
		hNewsg	    = 0;

	PROC ("PostNews");

	if (nodename == NULL) {
		err ("NodeName must be specified! Can't post news.", NULL);
		D (("exit -1\n"));
		return -1;
	}

	if (domain == NULL)
		domain = ".UUCP";

	sprintf (tfname, "T:post-%d", seq);
	fo = fopen (tfname, "w");
	if (fo == NULL) {
		err ("Cannot create '%s'\n", tfname);
		D (("exit -1\n"));
		return -1;
	}

	/*
	**  Create the news file.
	*/

	firsttime = 1;

	/*
	**  Check for user-specified headers
	*/
	while (fgets (TmpBuf, sizeof (TmpBuf), fi) && TmpBuf [0] != '\n') {
		if (firsttime) {
			firsttime = 0;
			if (STRIEQN (TmpBuf, HPath)) {
				/* path already there */
			}
			else {
				fprintf (fo, "%s%s\n", HPath, user);
				D (("Added %s%s\n", HPath, user));
			}
			hPath = 1;
		}

		if (STRIEQN (TmpBuf, HFrom))
			hFrom = 1;
		if (STRIEQN (TmpBuf, HDate))
			hDate = 1;
		if (STRIEQN (TmpBuf, HMsgid))
			hMsgid = 1;
		if (STRIEQN (TmpBuf, HOrg))
			hOrg = 1;
		if (STRIEQN (TmpBuf, HSubj))
			hSubj = 1;
		if (STRIEQN (TmpBuf, HNewsg))
			hNewsg = 1;
		fputs (TmpBuf, fo);
	}

	/*
	**  Other headers (-R option to postnews)
	**  (usually References, but can be any other stuff they want)
	*/

	if (rfi) {
		while (fgets (TmpBuf, sizeof (TmpBuf), rfi)) {
			if (STRIEQN (TmpBuf, HFrom))
				hFrom = 1;
			if (STRIEQN (TmpBuf, HDate))
				hDate = 1;
			if (STRIEQN (TmpBuf, HMsgid))
				hMsgid = 1;
			if (STRIEQN (TmpBuf, HOrg))
				hOrg = 1;
			if (STRIEQN (TmpBuf, HSubj))
				hSubj = 1;
			if (STRIEQN (TmpBuf, HNewsg))
				hNewsg = 1;
			fputs (TmpBuf, fo);
		}
	}

	if (hNewsg == 0) {
		/* Oops! */
		err ("No %sheader specified. Cannot post.\n", HNewsg);
		fclose (fo);
		remove (tfname);
		D (("exit -1\n"));
		return -1;
	}

	/*
	**  Add headers required but maybe not there yet
	*/

	if (hPath == 0) {
		fprintf (fo, "%s%s\n", HPath, user);
		D (("Added %s%s\n", HPath, user));
	}
	if (hFrom == 0) {
		fprintf (fo, "%s%s@%s%s (%s)\n", HFrom, user, nodename, domain, realname);
		D (("Added %s%s@%s%s (%s)\n", HFrom, user, nodename, domain, realname));
	}
	if (hMsgid == 0) {
		fprintf (fo, "%s<%s.%s@%s%s>\n", HMsgid, user, SeqToName (seq), nodename, domain);
		D (("Added %s<%s.%s@%s%s>\n", HMsgid, user, SeqToName (seq), nodename, domain));
	}
	if (hSubj == 0) {
		fprintf (fo, "%s<none>\n", HSubj);
		D (("Added %s<none>\n", HSubj));
	}
	if (hDate == 0) {
		struct tm
			*ut;
		time_t
			t = time (NULL);
		static char
			*dy [] = { "Sun", "Mon", "Tue", "Wed",
				   "Thu", "Fri", "Sat" },
			*mo [] = { "Jan", "Feb", "Mar", "Apr",
				   "May", "Jun", "Jul", "Aug",
				   "Sep", "Oct", "Nov", "Dec"};

		ut = localtime (&t);

		fprintf (fo, "%s%s, %d %s %02d %02d:%02d:%02d %s\n",
			HDate,
			dy [ut->tm_wday], ut->tm_mday,
			mo [ut->tm_mon], ut->tm_year % 100,
			ut->tm_hour, ut->tm_min, ut->tm_sec,
			timezone ? timezone : ""
		);
		D (("Added %s%s, %d %s %02d %02d:%02d:%02d %s\n",
			HDate,
			dy [ut->tm_wday], ut->tm_mday,
			mo [ut->tm_mon], ut->tm_year % 100,
			ut->tm_hour, ut->tm_min, ut->tm_sec,
			timezone ? timezone : ""
		));
	}

	if (hOrg == 0 && organization) {
		fprintf (fo, "%s%s\n", HOrg, organization);
		D (("Added %s%s\n", HOrg, organization));
	}

	/*
	**  remainder of message
	*/

	fprintf (fo, "\n");

	while (fgets (TmpBuf, sizeof (TmpBuf), fi)) {
		fputs (TmpBuf, fo);
	}

	fclose (fo);

	/*
	**  Post locally, also batches to other machines.
	*/

	if (rnewsnop && (*rnewsnop == 'y' || *rnewsnop == 'Y'))
		sprintf (TmpBuf, "%s %s", rnews, tfname);
	else
		sprintf (TmpBuf, "%s -p %s", rnews, tfname);

	D (("Calling '%s'\n", TmpBuf));
	rslt = system (TmpBuf);

	D (("result from system() = %ld\n", rslt));

	if (rslt) {
		ulog (-1, "RNews failure while posting news (rslt %ld, file '%s')\n", rslt, tfname);
	}
	else {
		remove (tfname);
	}

	if (autoBatch) {
		sprintf (TmpBuf, "%s", GetConfigProgram (BATCHNEWS));
		D (("Calling '%s'\n", TmpBuf));
		system (TmpBuf);
	}

	D (("exit, rslt %ld\n", rslt));
	return rslt;
}
